import json
import random
import string
import functools

from eth_utils.hexadecimal import remove_0x_prefix

from .data_type import SmartContract, EvmLogs, Tx
from .node import Node
from .constant import ASCOIN, SUCCESS, TxType
from ._utils.encode import encode_transaction_data as _encode_tx_data
from .evm_log import EvmLogParser


class Contract:
    """
    The primary entry point for working with smart contract.
    """

    def __init__(self, node: Node, address: str = None, c: SmartContract = None,
                 template_name: str = None, args: list = None):
        try:
            contract_template = node.get_contract_template(address=address, name=template_name)
        except:
            # create template firstly
            assert c is not None
            if template_name is None:
                template_name = ''.join(random.choices(string.ascii_letters + string.digits, k=5))
            tx_data = node.build_data_of_create_template(1, template_name, c.bytecode, c.abi, c.source)
            create_template_tx = node.call_write_function(
                contract_tx_data=tx_data, call_type=TxType.TEMPLATE
            ).broadcast()
            assert create_template_tx.check() is SUCCESS
            contract_template = node.get_contract_template(key=create_template_tx.id)

        if address is None:
            tx_data = node.build_data_of_deploy_contract(contract_template, args if args else [])
            tx = node.call_write_function(contract_tx_data=tx_data, call_type=TxType.CREATE)
            address = node._calc_contract_address(tx.transaction.vin, tx.transaction.vout)
            assert tx.broadcast().check() is SUCCESS

        self.template_name = contract_template.template_name
        self.address = address
        self.abi = contract_template.abi
        self.abi_json_str = json.dumps(self.abi)
        self.node: Node = node
        self.encode_tx_data = functools.partial(_encode_tx_data, contract_abi=self.abi)

    def __str__(self):
        return f"[address: {self.address}]"

    def __repr__(self):
        return self.__str__()

    def read(self, func_name, args=None):
        """
        call a view/pure function in the contract and return the execution result

        :param func_name: function name
        :param args: function arguments
        :return: the return value of the readonly function

        .. code-block:: python

            >>> from asimov import Contract, Node, AsimovSolc
            >>> node = Node("http://seed.asimov.tech")
            >>> contract = Contract(node, c=AsimovSolc.compile("/path/to/my/sources/example.sol")['Example'])
            >>> contract.read("readonly function name")
        """
        return self.node._call_readonly_function(
            contract_address=self.address,
            data=remove_0x_prefix(self.encode_tx_data(func_name, args=args)),
            func_name=func_name,
            abi=self.abi_json_str
        )

    def execute(self, func_name, args=None, asset_value=0, asset_type=ASCOIN, tx_fee_type=ASCOIN) -> Tx:
        """
        send a transaction to execute a function in the contract and return the transaction object :class:`~asimov.data_type.Tx`.
        Note the returned :class:`~asimov.data_type.Tx` object is in pending status.
        You need to check execution status constantly and wait it confirmed on chain.

        :param func_name: function name
        :param args: function arguments
        :param asset_value: the asset value to be send
        :param asset_type: the asset type to be send
        :param tx_fee_type: the transaction fee type
        :return: the :class:`~asimov.data_type.Tx` object

        .. code-block:: python

            >>> from asimov import Contract, Node, AsimovSolc
            >>> node = Node("http://seed.asimov.tech")
            >>> contract = Contract(node, c=AsimovSolc.compile("/path/to/my/sources/tutorial.sol")['Tutorial'])
            >>> tx = contract.execute("mint", [1000000])
            >>> tx.check()  # 1 / 0
        """
        return self.node.call_write_function(
            contract_address=self.address,
            params=args,
            func_name=func_name,
            abi=self.abi,
            asset_value=asset_value,
            asset_type=asset_type,
            tx_fee_type=tx_fee_type,
        ).broadcast()

    def vote(self, func_name, args=None, asset_value=0, asset_type=ASCOIN, tx_fee_type=ASCOIN) -> Tx:
        """
        send a transaction to vote on a contract and return the transaction object :class:`~asimov.data_type.Tx`
        Note the returned :class:`~asimov.data_type.Tx` object is in pending status.
        You need to check execution status constantly and wait it confirmed on chain.

        :param func_name: function name
        :param args: function arguments
        :param asset_value: the asset value to be send
        :param asset_type: the asset type to be send
        :param tx_fee_type: the transaction fee type
        :return: the :class:`~asimov.data_type.Tx` object

        .. code-block:: python

            >>> from asimov import Contract, Node, AsimovSolc
            >>> node = Node("http://seed.asimov.tech")
            >>> contract = Contract(node, c=AsimovSolc.compile("/path/to/my/sources/tutorial.sol")['Tutorial'])
            >>> contract.vote("vote", [1]).check()
        """
        return self.node.call_write_function(
            contract_address=self.address,
            params=args,
            func_name=func_name,
            abi=self.abi,
            asset_type=asset_type,
            asset_value=asset_value,
            call_type=TxType.VOTE,
            tx_fee_type=tx_fee_type
        ).broadcast()

    def fetch(self, tx_id) -> EvmLogs:
        """
        fetch the contract execution logs in the transaction

        :param tx_id: contract transaction id
        :return: the :class:`~asimov.data_type.EvmLogs` object
        """
        receipt = self.node._get_tx_receipt(tx_id)
        return EvmLogParser.parse(receipt['logs'], self.abi)



import json
import time
import copy
import math
from typing import Union

import requests
from eth_utils.address import remove_0x_prefix
from web3 import Web3

from .data_type import Account, Tx, ContractTemplate
from .account import AccountFactory
from . import error
from . import constant
from ._utils.encode import AsimovJsonEncoder, encode_transaction_data, encode_params
from ._utils.common import dict_add
from .transactions import Transaction


bytes_per_input = 148
gas_per_byte = 21


class Node:
    """
    A wrapped object for asimov node
    """
    def __init__(self, provider: str = None, private_key: str = None):
        self.session = requests.session()
        self.session.headers.update({"Content-type": "application/json"})
        self.provider = provider
        self.account: Account = AccountFactory.new(private_key)
        self.tx: Tx = None

    def __str__(self):
        return f"node[address:{self.address}]"

    def __repr__(self):
        return self.__str__()

    def set_rpc_server(self, url: str):
        """
        set rpc server url
        :param url: rpc server url
        """
        self.provider = url

    @staticmethod
    def _to_json(method: str, params: list = None):
        if not method.startswith(constant.RPC_PREFIX):
            method = constant.RPC_PREFIX + method
        data = {"id": int(time.time()), "jsonrpc": "2.0"}
        data.update({"method": method, "params": params})
        return data

    @property
    def address(self) -> str:
        """
        get the corresponding account address
        """
        return self.account.address

    @property
    def private_key(self):
        """
        get the account private key
        """
        return self.account.private_key

    def set_private_key(self, private_key: str):
        """
        set the account private key
        """
        self.account = AccountFactory.new(private_key)

    def call(self, method: str, args: list = None):
        """
        call asimov rpc service

        :param method: rpc function name
        :param args: rpc function arguments
        :return: the return value of the rpc function

        .. code-block:: python

            >>> from asimov import Node
            >>> node = Node("http://seed.asimov.tech")
            >>> node.call("getBlockChainInfo")
            {'chain': 'devnet',
             'blocks': 6841,
             'bestblockhash': '329a4289a46b5e6e6a7e63744338d63d9065c264a87f17c82bf13853806cc3ef',
             'mediantime': 1574942764,
             'pruned': False}
        """
        response = None
        try:
            if self.provider is None:
                raise Exception("provider is None")
            response = self.session.post(
                url=self.provider, data=json.dumps(self._to_json(method, args), cls=AsimovJsonEncoder))
            json_data = response.json()
            if "error" in json_data:
                raise error.RPCError(json_data['error']['message'])
            return json_data['result']
        except json.decoder.JSONDecodeError:
            if response:
                raise error.NetWorkError(response.content)
            raise error.JsonException()
        except KeyError as e:
            raise error.UnknownError(e)

    @staticmethod
    def create_contract_tx_output(address: str, amount: int, data: str, assets: str = constant.ASCOIN,
                                  contract_type=constant.TxType.CALL) -> dict:
        """
        create contract transaction output.
        in asimov, contract call is wrapped in transaction output.

        :param address: output address
        :param amount: output amount
        :param data: output data
        :param assets: output asset id in hex string format
        :param contract_type: contract call type, default is 'call'
        :return: output dict
        """
        return {
            "amount": amount,
            "address": address,
            "assets": assets,
            "contractType": contract_type,
            "data": data
        }

    @staticmethod
    def create_tx_output(address: str, amount: int, assets: str = constant.ASCOIN):
        """
        create transaction output.

        :param address: output address
        :param amount: output amount
        :param assets: output asset id in hex string format
        :return: output dict
        """
        return {
            "amount": amount,
            "address": address,
            "assets": assets
        }

    def balance(self, address: str = None, asset=constant.ASCOIN) -> Union[int, dict]:
        """
        get the balance of specific address by given asset type

        :param address: specific address, default is current node account address
        :param asset: asset type, will return balance for all asset types if none is given
        :return: balance of given asset type, or balance for all asset types
        """
        if address is None:
            address = self.address

        rst = self.call("getBalance", [address])
        if asset is not None:
            rst = [e for e in rst if e['asset'] == asset]
        if not rst:
            return 0
        for v in rst:
            v['value'] = int(v['value'])
        return rst[0]['value'] if asset is not None else rst

    def _get_tx_receipt(self, tx_id: str):
        return self.call("getTransactionReceipt", [tx_id])

    def check(self, tx_id: str) -> int:
        """
        If the transaction is a normal transaction, this function checks whether a transaction is confirmed on chain.
        If the transaction is a contract transaction, this function checks whether a transaction is confirmed on chain \
        and returns contract execution status.

        :param tx_id: transaction id
        :return: return 1 if the transaction is confirmed on chain and the execution result is success.

        .. code-block:: python

            >>> from asimov import Node, constant
            >>> node = Node("http://seed.asimov.tech", "0xafd29358a5ba9e2f5aac5cd5013a6830a99e34a68c469c78ab5f4c6f1d8c2a46")
            >>> tx = node.send("0x663bc0936166c07431ed04d7dc207eb7694e223ec4", asset_value=10)
            # wait tx on chain
            >>> assert tx.check() is constant.SUCCESS
        """
        assert self.wait_for_confirmation(tx_id) is True
        receipt = self._get_tx_receipt(tx_id)
        return int(receipt['status'], 16)

    def _get_utxo_in_page(self, address, asset, _from: int, count):
        """
        Get UTXO in page of a given address.

        :param address: the address to get UTXO from
        :param asset: asset type in hex string format
        :param from: start position of the page
        :param count: count of UTXO in the page
        :return: UTXO of specifc address by given asset type in page
        """

        rst = self.call("getUtxoInPage", [address, asset, _from, count])['utxos']
        return [item for item in rst if item['spendable'] is True]

    def _get_utxo(self, address: str, asset=constant.ASCOIN, amount=1):
        """
        Get UTXO with specific asset type and amount of a given address

        :param address: the address to get UTXO from
        :param asset: asset type, hex string format, default value is '000000000000'
        :param amount: asset amount, default is 1
        """
        amount = max([amount, 1])
        utxos = []
        utxo_pool = []
        current_amount = 0
        idx = 0
        while current_amount < amount:
            _utxos = self._get_utxo_in_page(address, asset, idx, 1000)
            if len(_utxos) == 0:
                raise error.NotEnoughMoney(utxos)
            _utxos = [utxo for utxo in _utxos if (utxo['txid'], utxo['vout']) not in utxo_pool]
            total_amount = sum([utxo['amount'] for utxo in _utxos])
            if total_amount + current_amount >= amount:
                for utxo in _utxos:
                    utxos.append(utxo)
                    current_amount += utxo['amount']
                    if current_amount >= amount:
                        return utxos
            else:
                utxos.extend(_utxos)
                utxo_pool.extend([(utxo['txid'], utxo['vout']) for utxo in _utxos])
                current_amount += total_amount
                idx += 1000

    def _send_raw_trx(self, signed_tx: str):
        """send signed raw transaction"""

        return self.call("sendRawTransaction", [signed_tx])

    def _get_raw_tx(self, tx_id: str, need_detail=False, need_extra=False):
        """get raw transaction detail by transaction id"""

        return self.call("getRawTransaction", [tx_id, need_detail, need_extra])

    def wait_for_confirmation(self, tx_id, confirm_num=1, timeout=60) -> bool:
        """
        wait for the transaction to be confirmed on chain

        :param tx_id: transaction id
        :param confirm_num: confirmed block count, default is 1
        :param timeout: time out length, default is 60 seconds
        :return: true if confirmed, false otherwise
        """
        end_time = time.time() + timeout
        while time.time() < end_time:
            rst = self._get_raw_tx(tx_id, True)
            if rst.get("confirmations", 0) >= confirm_num:
                return True
            time.sleep(1)
        return False

    def _get_best_block(self):
        """get the best block"""
        return self.call("getBestBlock")

    @property
    def current_height(self) -> int:
        """
        get the current height of chain
        """
        return self._get_best_block()['height']

    def _calc_contract_address(self, inputs: list, outputs: list):
        """
        calculate contract address from transaction inputs and outputs

        :param inputs: transaction inputs
        :param outputs: transaction outputs
        :return: contract address
        """
        outputs = copy.deepcopy(outputs)
        for output in outputs:
            output['amount'] = str(output['amount'])
        return self.call("calculateContractAddress", [inputs, outputs])['0']

    def get_contract_template(self, address: str = None, key: str = None, name: str = None) -> ContractTemplate:
        """
        get contract template object according to address, key or name

        :param address: contract address
        :param key: template key (template id)
        :param name: template name
        :return: the :class:`~asimov.data_type.ContractTemplate` object
        """
        if address:
            rst = self.call("getContractTemplate", [address])
            rst = self.call("getContractTemplateInfoByName", [1, rst['template_name']])
        elif key:
            rst = self.call("getContractTemplateInfoByKey", [key])
        elif name:
            rst = self.call("getContractTemplateInfoByName", [1, name])
        else:
            raise error.UnknownError()
        return ContractTemplate(rst['template_name'], rst['category'], rst['source'],
                                json.loads(rst['abi']), rst['byte_code'])

    def _call_readonly_function(self, contract_address: str, data: str, func_name: str,
                                abi: str, caller_address: str = None):
        """
        call a readonly function in a contract. a readonly function is marked as 'pure' or 'view'.

        :param contract_address: the contract to call
        :param data: binary data
        :param func_name: the function name to call
        :param abi: abi of the contract
        :param caller_address: the address to make the contract call
        """
        if caller_address is None:
            caller_address = self.address
        return self.call("callReadOnlyFunction", [caller_address, contract_address, data, func_name, abi])

    def _select_utxo(self, assets: dict) -> dict:
        """
        select UTXO according to given parameters
        """

        rst = dict()
        utxos = []

        for k, v in assets.items():
            if v == 0:
                continue
            _utxos = self._get_utxo(self.address, k, v)
            for _utxo in _utxos:
                _utxo['signed_key'] = self.account
            rst[k] = sum([_utxo['amount'] for _utxo in _utxos])
            utxos.extend(_utxos)
        rst['utxos'] = utxos

        return rst

    def _select_vote_utxo(self, vote_value: int, vote_asset_type: str, fees: dict) -> dict:
        """select UTXO for vote transaction"""
        assets = copy.copy(fees)
        if vote_value == 0:
            balance_of_vote_asset_type = self.balance(address=self.address, asset=vote_asset_type)
            if fees.get(vote_asset_type, 0) > balance_of_vote_asset_type:
                raise error.NotEnoughMoney(
                    f"need {assets[vote_asset_type]}, but only have {balance_of_vote_asset_type} of {vote_asset_type}")
            else:
                assets[vote_asset_type] = balance_of_vote_asset_type
        else:
            assets[vote_asset_type] = assets.get(vote_asset_type, 0) + vote_value
        return self._select_utxo(assets)

    def _build_transfer(
            self, address, asset_value, asset_type=constant.ASCOIN,
            tx_fee_type=constant.ASCOIN, gas_price=constant.DEFAULT_GAS_PRICE
    ) -> Transaction:
        """
        build a transaction
        """
        def __build_outputs(_select_rst: dict, _fee_value: int, _fee_type: str) -> list:
            needed_assets = dict_add({asset_type: asset_value}, {_fee_type: _fee_value})
            _outputs = [{
                "address": address,
                "amount": asset_value,
                "assets": asset_type
            }]
            # 找零 output
            for _asset_type in _select_rst:
                if _select_rst[_asset_type] > needed_assets[_asset_type]:
                    _outputs.append({
                        "address": self.address,
                        "amount": _select_rst[_asset_type] - needed_assets[_asset_type],
                        "assets": _asset_type
                    })
            return [item for item in _outputs if item['amount'] > 0]

        tx_fee_value = 0
        while True:
            select_rst = self._select_utxo(dict_add({asset_type: asset_value}, {tx_fee_type: tx_fee_value}))
            inputs = select_rst.pop('utxos')
            outputs = __build_outputs(select_rst, tx_fee_value, tx_fee_type)
            tx = Transaction(inputs, outputs)
            gas = self.estimate_gas(tx.sign().to_hex(), inputs)
            tx_fee_value = math.ceil(gas * gas_price)
            if select_rst.get(tx_fee_type, 0) >= (asset_value if asset_type == tx_fee_type else 0) + tx_fee_value:
                outputs = __build_outputs(select_rst, tx_fee_value, tx_fee_type)
                break
            tx_fee_value += math.ceil(bytes_per_input * gas_per_byte * gas_price)
        return Transaction(inputs, outputs, gas_limit=gas)

    def send(self, address, asset_value: int, asset_type=constant.ASCOIN, tx_fee_type=constant.ASCOIN) -> Tx:
        """
        send a normal transaction on asimov chain and return the transaction object :class:`~asimov.data_type.Tx`

        :param address: target address
        :param asset_value: asset value to send
        :param asset_type: asset type to send
        :param tx_fee_type: transaction fee type
        :return: the :class:`~asimov.data_type.Tx` object

        .. code-block:: python

            >>> from asimov import Node, constant
            >>> node = Node("http://seed.asimov.tech", "0x98ca5264f6919fc12536a77c122dfaeb491ab01ed657c6db32e14a252a8125e3")
            >>> node.send("0x663bc0936166c07431ed04d7dc207eb7694e223ec4", asset_value=10, asset_type=constant.ASCOIN, tx_fee_type=constant.ASCOIN)
            [id: 91c4645bcf3680c699a591632cd8769abe2973fd2de70081a6752d9781f2801b]
        """
        if asset_value < 1:
            raise error.InvalidParams(f"value should be larger than 1, got {asset_value}")
        return Tx(self, self._build_transfer(address, asset_value, asset_type, tx_fee_type)).broadcast()

    def call_write_function(
            self, func_name: str = None, params: tuple = None, abi=None, contract_address: str = constant.NullAddress,
            contract_tx_data=None, call_type=constant.TxType.CALL, asset_value=0, asset_type=constant.ASCOIN,
            tx_fee_type=constant.ASCOIN, gas_price=constant.DEFAULT_GAS_PRICE, corrected_gas=50000
    ) -> Tx:
        """
        send a transaction to execute a method in the contract

        :param func_name: function name to be called in the contract
        :param params: call parameters
        :param abi: contract abi
        :param contract_address: contract address
        :param contract_tx_data: binary data of the transaction
        :param call_type: call type
        :param asset_value: asset value to send
        :param asset_type: asset type to send
        :param tx_fee_type: transaction fee type
        :param gas_price: gas price
        :param corrected_gas: adjusted gas value
        :return: the :class:`~asimov.data_type.Tx` object
        """
        if call_type in (constant.TxType.TEMPLATE, constant.TxType.CREATE):
            assert contract_tx_data is not None
        if call_type in (constant.TxType.CALL, constant.TxType.VOTE) and contract_tx_data is None:
            contract_tx_data = remove_0x_prefix(
                encode_transaction_data(fn_identifier=func_name, contract_abi=abi, args=params)
            )
        contract_output = self.create_contract_tx_output(
            address=contract_address,
            amount=asset_value,
            data=contract_tx_data,
            assets=asset_type,
            contract_type=call_type
        )

        # estimate gas
        tx_fee_value = 0
        while True:
            if call_type == constant.TxType.VOTE:
                select_rst = self._select_vote_utxo(0, asset_type, {tx_fee_type: tx_fee_value})
            else:
                need_assets = dict_add({asset_type: asset_value}, {tx_fee_type: tx_fee_value})
                need_assets[tx_fee_type] = max([need_assets[tx_fee_type], 1])
                select_rst = self._select_utxo(need_assets)
            inputs = select_rst['utxos']
            outputs = [contract_output]
            if asset_type == tx_fee_type:
                outputs.append(self.create_tx_output(
                    self.address,
                    select_rst.get(asset_type, 0) - asset_value - tx_fee_value,
                    asset_type
                ))
            else:
                outputs.extend([
                    self.create_tx_output(self.address, select_rst.get(asset_type, 0) - asset_value, asset_type),
                    self.create_tx_output(self.address, select_rst.get(tx_fee_type, 0) - tx_fee_value, tx_fee_type)
                ])
            valid_outputs = [item for item in outputs if not (len(item.get("data", "")) == 0 and item['amount'] <= 0)]
            tx = Transaction(inputs, valid_outputs)
            gas = self.estimate_gas(tx.sign().to_hex(), inputs, corrected_gas)
            tx_fee_value = math.ceil(gas * gas_price)
            min_value = tx_fee_value
            if asset_type == tx_fee_type:
                min_value += asset_value
            if select_rst.get(tx_fee_type, 0) >= min_value:
                break
            tx_fee_value += math.ceil(bytes_per_input * gas_per_byte * gas_price)

        outputs = [contract_output]
        if asset_type == tx_fee_type:
            outputs.append(
                self.create_tx_output(self.address, select_rst.get(asset_type, 0) - asset_value - tx_fee_value, asset_type)
            )
        else:
            outputs.extend([
                self.create_tx_output(self.address, select_rst.get(asset_type, 0) - asset_value, asset_type),
                self.create_tx_output(self.address, select_rst.get(tx_fee_type, 0) - tx_fee_value, tx_fee_type)
            ])
        valid_outputs = [item for item in outputs if not (len(item.get("data", "")) == 0 and item['amount'] <= 0)]
        return Tx(self, Transaction(select_rst['utxos'], valid_outputs, gas_limit=gas), is_contract_tx=True)

    def estimate_gas(self, tx_hex: str, inputs: list, corrected_value=50000):
        """
        estimate transaction execution gas cost
        :param tx_hex: transaction in hex format
        :param inputs: UTXO used in transaction
        :param corrected_value: corrected gas value
        :return:
        """
        rst = self.call("runTransaction", [tx_hex, inputs])
        return rst['gasUsed'] + corrected_value

    @staticmethod
    def build_data_of_deploy_contract(contract_template: ContractTemplate, params: list) -> str:
        """
        binary data for contract deployment transaction

        :param contract_template: the :class:`~asimov.data_type.ContractTemplate` object
        :param params: parameters of contract constructor function
        :return: transaction data
        """
        category_hex_str = remove_0x_prefix(hex(contract_template.category)).zfill(4)
        template_name_hex = bytes(contract_template.template_name, 'utf-8').hex()
        template_name_length_hex = str(len(contract_template.template_name)).zfill(8)

        params_hex = encode_params(contract_template.abi, None, constant.ContractFunType.CONSTRUCTOR, params)
        return category_hex_str + template_name_length_hex + template_name_hex + params_hex

    @staticmethod
    def build_data_of_create_template(category, name, hex_code, abi, source="solidity source code") -> str:
        """
        binary data for template creation transaction

        :param category: template category
        :param name: template name
        :param hex_code: transaction data in hex type
        :param abi: template abi
        :param source: template source code
        :return: transaction data
        """
        MAX = 0xffff
        if category >= MAX:
            category = MAX
        name_bytes: bytes = Web3.toBytes(text=name)
        bytecode_bytes = Web3.toBytes(hexstr=hex_code)
        if not isinstance(abi, str):
            abi = json.dumps(abi)
        abi_bytes: bytes = Web3.toBytes(text=abi)
        source_bytes: bytes = Web3.toBytes(text=source)
        return "".join([
            category.to_bytes(2, 'big', signed=False).hex(),
            len(name_bytes).to_bytes(4, 'big', signed=False).hex(),
            len(bytecode_bytes).to_bytes(4, 'big', signed=False).hex(),
            len(abi_bytes).to_bytes(4, 'big', signed=False).hex(),
            len(source_bytes).to_bytes(4, 'big', signed=False).hex(),
            name_bytes.hex(),
            hex_code,
            abi_bytes.hex(),
            source_bytes.hex()
        ])



from datetime import datetime


class _BaseException(Exception):
    def __repr__(self):
        return f"[{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] {self.args}"

    def __str__(self):
        return self.__repr__()


class NoUtxoError(_BaseException):
    """no available utxo"""


class NotEnoughMoney(_BaseException):
    """not enough money"""


class JsonException(_BaseException):
    """json"""


class RPCError(_BaseException):
    """rpc response error"""


class NoAvailableKey(_BaseException):
    """no available key"""


class InvalidPrivateKey(_BaseException):
    """invalid private key"""


class CompileError(_BaseException):
    """compile error"""


class CompileTimeout(_BaseException):
    """compile timeout"""


class NetWorkError(_BaseException):
    """network error"""


class UnknownError(_BaseException):
    """Unknown"""


class UnSupportSolidityImportType(_BaseException):
    """solidity import error"""


class InvalidParams(_BaseException):
    """invalid parameters"""


class InvalidTxType(_BaseException):
    """invalid tx type"""



import secrets
import hashlib
import codecs

from fastecdsa import (
    keys,
    curve
)
from eth_utils import (
    add_0x_prefix,
    remove_0x_prefix,
)
from web3 import Web3
from bitcointx.core import script

from .data_type import Account
from .constant import AddressType, AsimovOpCode


class PrivateKeyFactory:
    """Private key generator"""

    _KEY_BYTES = 32
    _CURVE_ORDER = int('FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141', 16)

    @classmethod
    def generate_key(cls) -> str:
        """
        Create a new private key and return.

        :return: a private key in string form.

        .. code-block:: python

            >>> from asimov import PrivateKeyFactory
            >>> PrivateKeyFactory.generate_key()
            'bba692e559fda550d0157669b101bafddb23e7f57aeeb5cef5494e7a41a1f056'

            # note that the key generated above is a random value, it will be different in your try
        """
        big_int = secrets.randbits(cls._KEY_BYTES * 8)
        big_int %= cls._CURVE_ORDER - 1
        big_int += 1
        key = hex(big_int)[2:].zfill(cls._KEY_BYTES * 2)
        return key


class AccountFactory:
    """
    The primary entry point for working with Asimov private keys.

    It does **not** require a connection to an Asimov node.
    """

    @classmethod
    def generate_address(cls, private_key: str) -> str:
        """
        create an asimov address with the given private key

        :param private_key: the private key
        :return: an hex address generated from private key, without 0x prefix.

        .. code-block:: python

            >>> from asimov import AccountFactory
            >>> AccountFactory.generate_address("bba692e559fda550d0157669b101bafddb23e7f57aeeb5cef5494e7a41a1f056")
            '66c17b951f0c85b860c9f7f0d811c77ea78f2d2e3a'
        """
        public_key = cls.private2compressed_public(remove_0x_prefix(private_key))
        address = cls.__public2address(public_key)
        return address

    @classmethod
    def private2account(cls, private_key: str) -> Account:
        """
        convert private key to account and return as a :class:`~asimov.data_type.Account`

        :param private_key: the private key
        :return: an :class:`~asimov.data_type.Account` object with address, private key and public key.

        .. code-block:: python

            >>> from asimov import AccountFactory
            >>> account = AccountFactory.private2account("bba692e559fda550d0157669b101bafddb23e7f57aeeb5cef5494e7a41a1f056")
            >>> account.address
            '0x66c17b951f0c85b860c9f7f0d811c77ea78f2d2e3a'
            >>> account.private_key
            '0xbba692e559fda550d0157669b101bafddb23e7f57aeeb5cef5494e7a41a1f056'
            >>> account.public_key
            b'023a68576342553357f042c6ede12bd3ed01cb61ad39848908883cab93f66c7601'
        """
        public_key = cls.private2compressed_public(remove_0x_prefix(private_key))
        address = cls.__public2address(public_key)
        return Account(add_0x_prefix(private_key), add_0x_prefix(address), public_key)

    @classmethod
    def __private2public(cls, private_key: str) -> bytes:
        point = keys.get_public_key(Web3.toInt(hexstr=private_key), curve.secp256k1)
        public_key = '%s%s' % ('%064x' % int(point.x), '%064x' % int(point.y))
        return public_key.encode()

    @staticmethod
    def __add_bitcoin_byte(v: bytes) -> bytes:
        bitcoin_byte = b'04'
        return bitcoin_byte + v

    @classmethod
    def private2public(cls, private_key: str) -> bytes:
        """
        generate public key from private key

        :param private_key: private key
        :return: public key in bytes format

        .. code-block: python

            >>> from asimov import AccountFactory
            >>> AccountFactory.private2public("0xbba692e559fda550d0157669b101bafddb23e7f57aeeb5cef5494e7a41a1f056")
            >>> b'043a68576342553357f042c6ede12bd3ed01cb61ad39848908883cab93f66c76016fb60b2b472d6caf316c699cb38f61d5daef3792402461ddc449a18b0fc8ee32'
        """
        key_hex = cls.__private2public(private_key)
        return cls.__add_bitcoin_byte(key_hex)

    @classmethod
    def private2compressed_public(cls, private_key: str) -> bytes:
        """
        generate compressed public key from private key

        :param private_key: private key
        :return: compressed public key in bytes format

        .. code-block: python

            >>> from asimov import AccountFactory
            >>> AccountFactory.private2compressed_public("0xbba692e559fda550d0157669b101bafddb23e7f57aeeb5cef5494e7a41a1f056")
            >>> b'023a68576342553357f042c6ede12bd3ed01cb61ad39848908883cab93f66c7601'
        """
        key_hex = cls.__private2public(private_key)
        # get x from key (first half)
        key_string = str(key_hex, 'utf8')
        key_half = key_hex[:(len(key_hex) // 2)]
        # Add bitcoin byte: 0x02 if the last digit is even, 0x03 if the last digit is odd
        last_byte = int(key_string[-1], 16)
        btc_byte = b'02' if last_byte % 2 == 0 else b'03'
        public_key = btc_byte + key_half
        return public_key

    @classmethod
    def __public2address(cls, public_key: bytes) -> str:
        publick_key_bytes = codecs.decode(public_key, 'hex')
        # Run SHA256 for the public key
        sha256_bpk_digest = hashlib.sha256(publick_key_bytes).digest()
        # Run ripemd160 for the SHA256
        ripemd160_bpk = hashlib.new('ripemd160')
        ripemd160_bpk.update(sha256_bpk_digest)
        ripemd160_bpk_digest = ripemd160_bpk.digest()

        # asimov logic
        public_key_hash_addr_id = b'f'
        address_length = 21
        digest_length = len(ripemd160_bpk_digest)
        if digest_length > address_length - 1:
            ripemd160_bpk_digest = ripemd160_bpk_digest[digest_length + 1 - address_length:]
        ripemd160_bpk_digest = public_key_hash_addr_id + ripemd160_bpk_digest
        return str(codecs.encode(ripemd160_bpk_digest, 'hex'), 'utf8')

    @classmethod
    def new(cls, private_key=None) -> Account:
        """
        create a new account from the given private key and return as a :class:`~asimov.data_type.Account`

        :param private_key: private key, if None, a new private key will be generated
        :return: an :class:`~asimov.data_type.Account` object

        .. code-block:: python

            >>> from asimov import AccountFactory
            >>> account = AccountFactory.new("bba692e559fda550d0157669b101bafddb23e7f57aeeb5cef5494e7a41a1f056")
            >>> account.address
            '0x66c17b951f0c85b860c9f7f0d811c77ea78f2d2e3a'
            >>> account.private_key
            '0xbba692e559fda550d0157669b101bafddb23e7f57aeeb5cef5494e7a41a1f056'
            >>> account.public_key
            b'023a68576342553357f042c6ede12bd3ed01cb61ad39848908883cab93f66c7601'
        """
        if private_key is None:
            private_key = PrivateKeyFactory.generate_key()
        return cls.private2account(private_key)


class Address:
    def __init__(self, address: str):
        self.address = address

    def __str__(self):
        return self.address

    def to_str(self):
        return self.address

    @property
    def type(self):
        return Web3.toInt(hexstr=remove_0x_prefix(self.address)[:2])

    @property
    def is_pay_to_contract_hash(self):
        return self.type == AddressType.ContractHash

    @property
    def is_pay_to_pub_hash(self):
        return self.type == AddressType.PubKeyHash

    def to_script_pub_key(self) -> str:
        return script.CScript([
            script.OP_DUP, script.OP_HASH160,
            Web3.toBytes(hexstr=self.address), AsimovOpCode.OP_IFLAG_EQUALVERIFY, script.OP_CHECKSIG
        ]).hex()

    @staticmethod
    def to_create_contract_hash_script():
        return script.CScript(AsimovOpCode.OP_CREATE.to_bytes(1, 'little')).hex()

    @staticmethod
    def to_create_template_hash_script():
        return script.CScript(AsimovOpCode.OP_TEMPLATE.to_bytes(1, 'little')).hex()

    def to_create_vote_hash_script(self):
        return ''.join([
            script.CScript(AsimovOpCode.OP_VOTE.to_bytes(1, 'little')).hex(),
            script.CScript(AsimovOpCode.OP_DATA_21.to_bytes(1, 'little')).hex(),
            remove_0x_prefix(Web3.toHex(hexstr=self.address))
        ])

    def to_contract_hash_script(self):
        return script.CScript(AsimovOpCode.OP_CALL.to_bytes(1, 'little')).hex() + \
               len(Web3.toBytes(hexstr=self.address)).to_bytes(1, 'little', signed=False).hex() + \
               remove_0x_prefix(Web3.toHex(hexstr=self.address))



from bitcointx.core.script import CScriptOp

# Asim is the system asset of Asimov blockchain
ASCOIN = "000000000000000000000000"

SUCCESS = 1
FAILED = 0

RPC_PREFIX = 'asimov_'


# Xin is the smallest asset unit in Asimov, which is the same as satoshi in Bitcoin
# Coin = 100,000,000 Xin
XIN = 1
COIN = 100_000_000 * XIN

DEFAULT_GAS_PRICE = 0.1

NullAddress = "0x660000000000000000000000000000000000000000"


class TxType:
    """
    There are different call types when interacting with contract in a transaction

    #. CREATE deploy a new contract
    #. CALL call a contract function
    #. TEMPLATE submit a new template to template warehouse
    #. VOTE vote to a contact function
    """
    CREATE = "create"
    CALL = "call"
    TEMPLATE = "template"
    VOTE = "vote"


class AsimovOpCode:
    """new opcode added in asimov"""
    OP_DATA_21 = CScriptOp(21)
    OP_TEMPLATE = CScriptOp(192)
    OP_CREATE = CScriptOp(193)
    OP_CALL = CScriptOp(194)
    OP_SPEND = CScriptOp(195)
    OP_IFLAG_EQUAL = CScriptOp(196)
    OP_IFLAG_EQUALVERIFY = CScriptOp(197)
    OP_VOTE = CScriptOp(198)


class AddressType:
    """
    address types

    #. 0x66 normal account
    #. 0x63 contract
    """
    PubKeyHash = 0x66
    ContractHash = 0x63


class ContractFunType:
    """function types which follow EVM standards"""
    CONSTRUCTOR = "constructor"
    FUN = "function"
    FALLBACK = "fallback"
    EVENT = "event"


class UtxoSelectPolicy:
    NORMAL = 'normal'
    VOTE = 'vote'




from collections import namedtuple
from pprint import pformat

from eth_utils.hexadecimal import remove_0x_prefix
from .constant import SUCCESS, FAILED


def __smart_contract_repr(c):
    return pformat({
        "abi": c.abi,
        "bytecode": c.bytecode,
        "address": c.address,
        "source": c.source
    })


SmartContract = namedtuple("SmartContract", ("source", "abi", "bytecode", "address"))
SmartContract.__new__.__defaults__ = (None,) * 4
SmartContract.__repr__ = __smart_contract_repr
SmartContract.__doc__ = "compiled contract"
SmartContract.source.__doc__ = "contract source code"
SmartContract.abi.__doc__ = "contract abi"
SmartContract.bytecode.__doc__ = "contract bytecode"
SmartContract.address.__doc__ = "contract address on chain"

ContractTemplate = namedtuple("ContractTemplate", ("template_name", "category", "source", "abi", "byte_code"))
ContractTemplate.__new__.__defaults__ = (None,) * 5
ContractTemplate.__doc__ = "contract template"
ContractTemplate.template_name.__doc__ = "template name"
ContractTemplate.category.__doc__ = "template category"
ContractTemplate.source.__doc__ = "template source code"
ContractTemplate.abi.__doc__ = "template abi"
ContractTemplate.byte_code.__doc__ = "template bytecode"


class Account:
    """
    Asimov account, consists of private key, public key and address
    """
    def __init__(self, private_key=None, address=None, public_key=None):
        self.private_key = private_key
        self.public_key = public_key
        self.address = address

    def __str__(self):
        return f"[{self.private_key}, {self.address}]"

    def __repr__(self):
        return self.__str__()


class Tx:
    """
    The primary entry point for working with transaction object.
    """

    def __init__(self, node, transaction, _id=None, is_contract_tx=False):
        self.node = node
        self.transaction = transaction
        self.signed_hex = transaction.sign().to_hex()
        self._id = _id
        self.is_contract_tx = is_contract_tx

    def __repr__(self):
        return f"[id: {self.id}]"

    @property
    def id(self) -> str:
        """
        get the transaction id
        """
        return self._id

    @id.setter
    def id(self, _id: str):
        """
        set the transaction id
        """
        self._id = _id

    def check(self) -> int:
        """
        check whether a normal transaction is confirmed on chain, or a contract call is successful or not
        
        :return: 1 if the transaction is confirmed on chain, or the contract call is successful
        """
        if self.is_contract_tx:
            return self.node.check(self.id)
        else:
            return SUCCESS if self.node.wait_for_confirmation(self.id) else FAILED

    def broadcast(self):
        """
        broadcast the transaction
        :return:
        """
        self._id = self.node._send_raw_trx(self.signed_hex)
        return self


class Asset:
    """
    The primary entry point for working with asset on asimov chain.

    Asimov asset consists of 3 parts

    #. asset_type, 4 bytes long, each bit contains an asset property. For now, the first bit is used to determine whether an asset is divisible and the second bit is used to determine whether the asset is restricted.
    #. org_id, 4 bytes long organization id, system wide unique id assigned to organization when registering to asimov platform.
    #. asset_index, 4 bytes long asset index in organization, the assigning rule is determined by the organization itself.
    """

    @staticmethod
    def asset_wrapper(asset_type, org_id, asset_index) -> int:
        """
        asimov asset wrapper

        :param asset_type: asset type.
        :param org_id: organization id.
        :param asset_index: asset index.
        :return: asimov asset id in int format

        .. code-block:: python

            >>> from asimov import Asset
            >>> Asset.asset_wrapper(0, 1, 1)
            4294967297
        """
        return (asset_type << 64) + (org_id << 32) + asset_index

    @staticmethod
    def asset2str(asset: int) -> str:
        """
        convert asset id from int to hex string without 0x
        :param asset: asset id in int format
        :return: asset id in hex string format

        .. code-block:: python

            >>> from asimov import Asset
            >>> Asset.asset2str(4294967297)
            '000000000000000100000001'
        """
        return remove_0x_prefix(hex(asset)).zfill(24)

    def __init__(self, contract, asset_type: int, asset_index: int):
        self.contract = contract
        self.org_id: int = self.contract.read("orgId")
        self.asset_type: int = asset_type
        self.asset_index: int = asset_index

    def __repr__(self):
        return f"[contract address: {self.contract.address}, asset: {self.asset_id_str} / {self.asset_id_int}]"

    @property
    def asset_id_int(self) -> int:
        """
        get the asset id in int format
        """
        return self.asset_wrapper(self.asset_type, self.org_id, self.asset_index)

    @property
    def asset_id_str(self) -> str:
        """
        get the asset id in hex string format
        """
        return self.asset2str(self.asset_id_int)


class EvmLog:
    """
    contract execution log
    """
    def __init__(self, raw_log):
        self.address = raw_log['address']
        self.block_hash = raw_log['blockHash']
        self.transaction_hash = raw_log['transactionHash']
        self.data = raw_log['data']
        self.topics: [str] = raw_log['topics']


class EvmLogs(list):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def to_dict(self) -> dict:
        """
        convert evm logs from list to dict, which may lose the logs of the same name
        :return: evm logs in dict format
        """
        return {item['name']: item['args'] for item in self}



import struct
from copy import deepcopy
from io import StringIO, BytesIO

from web3 import Web3
from eth_utils import remove_0x_prefix
from bitcointx.wallet import CBitcoinSecret
from bitcointx.core import script, b2lx, Hash
from bitcointx.core.serialize import VarIntSerializer

from .constant import TxType
from .data_type import Account
from .account import Address


DEFAULT_SEQUENCE = 0xffffffff


def VarIntSerializeSize(val: int) -> int:
    if val < 0xfd:
        return 1
    elif val <= (1 << 16 - 1):
        return 3
    elif val <= (1 << 32 - 1):
        return 5
    return 9


class TxInput:
    def __init__(self, vin: dict):
        self.prev_tx_id = vin['txid']
        self.vout = vin['vout']
        self.sequence = DEFAULT_SEQUENCE
        self.scriptPubKey = vin['scriptPubKey']
        self.sig_script = None
        self.signed_key: Account = vin.get('signed_key')

    def write_buffer(self, buf):
        buf.write(self.prev_tx_id)
        buf.write(b2lx((self.vout).to_bytes(4, 'big', signed=False)))
        if self.sig_script:
            buf.write(VarIntSerializer.serialize(len(Web3.toBytes(hexstr=self.sig_script))).hex())
            buf.write(self.sig_script)
        else:
            buf.write(b'\x00'.hex())
        buf.write(b2lx((self.sequence).to_bytes(4, 'big', signed=False)))
        return buf

    def serialize_size(self) -> int:
        # Outpoint Hash 32 bytes + Outpoint Index 4 bytes + Sequence 4 bytes +
        # serialized varint size for the length of SignatureScript +
        # SignatureScript bytes
        length = len(self.sig_script) / 2 if self.sig_script else 0
        return 40 + length + VarIntSerializeSize(length)


class TxOutput:
    def __init__(self, output: dict):
        self.address = Address(output['address'])
        self.amount = output['amount']
        self.assets = output['assets']
        self.contract_type = output.get("contractType")
        self.data = output.get("data")
        self.pk_script = None
        if self.data:
            self.data = remove_0x_prefix(self.data)
            if self.contract_type == TxType.CALL:
                self.pk_script = self.address.to_contract_hash_script()
            elif self.contract_type == TxType.TEMPLATE:
                self.pk_script = self.address.to_create_template_hash_script()
            elif self.contract_type == TxType.CREATE:
                self.pk_script = self.address.to_create_contract_hash_script()
            elif self.contract_type == TxType.VOTE:
                self.pk_script = self.address.to_create_vote_hash_script()
            else:
                raise TypeError(f"unknown contract type: {self.contract_type}")
        else:
            if self.address.is_pay_to_contract_hash:
                self.pk_script = self.address.to_contract_hash_script()
            else:
                self.pk_script = self.address.to_script_pub_key()

    def write_buffer(self, buf):
        buf.write(b2lx(struct.pack(">q", self.amount)))
        buf.write(VarIntSerializer.serialize(len(Web3.toBytes(hexstr=self.pk_script))).hex())
        buf.write(self.pk_script)
        if self.assets:
            buf.write(VarIntSerializer.serialize(len(Web3.toBytes(hexstr=self.assets))).hex())
            buf.write(self.assets)
        else:
            buf.write(b'\x00'.hex())
        if self.data:
            buf.write(VarIntSerializer.serialize(len(Web3.toBytes(hexstr=self.data))).hex())
            buf.write(self.data)
        else:
            buf.write(b'\x00'.hex())
        return buf

    def serialize_size(self) -> int:
        # Value 8 bytes + Assets 12 byte
        # + serialized varint size for the length of PkScript +
        # PkScript bytes.
        # + serialized varint size for the length of Data +
        # Data bytes.
        pk_script_length = len(self.pk_script) / 2 if self.pk_script else 0
        data_length = len(self.data) / 2 if self.data else 0
        return 20 + VarIntSerializeSize(pk_script_length) + pk_script_length + VarIntSerializeSize(data_length) + data_length


class Transaction:
    def __init__(self, vin=(), vout=(), nLockTime=0, nVersion=1, gas_limit=0):
        if not (0 <= nLockTime <= 0xffffffff):
            raise ValueError('CTransaction: nLockTime must be in range 0x0 to 0xffffffff; got %x' % nLockTime)
        self.vin = vin
        self.vout = vout
        self.lock_time = nLockTime
        self.version = nVersion
        self.gas_limit = gas_limit
        self.inputs = [TxInput(_vin) for _vin in vin]
        self.outputs = [TxOutput(_vout) for _vout in vout]

    def to_buffer_writer(self) -> StringIO:
        buffer = StringIO()
        buffer.write(b2lx((self.version).to_bytes(4, 'big', signed=False)))
        buffer.write(b2lx(VarIntSerializer.serialize(len(self.inputs))))
        for _input in self.inputs:
            _input.write_buffer(buffer)
        buffer.write(b2lx(VarIntSerializer.serialize(len(self.outputs))))
        for _output in self.outputs:
            _output.write_buffer(buffer)
        buffer.write(b2lx((self.gas_limit).to_bytes(4, 'big', signed=False)))
        buffer.write(b2lx((self.lock_time).to_bytes(4, 'big', signed=True)))
        return buffer

    def to_hex(self) -> str:
        return self.to_buffer_writer().getvalue()

    def sign(self, sig_type=script.SIGHASH_ALL):
        for idx, _input in enumerate(self.inputs):
            assert _input.signed_key is not None
            _input.sig_script = AsimovScript.sign(self, _input.signed_key, idx, _input.scriptPubKey, sig_type).hex()
        return self

    def serialize_size(self) -> int:
        # Version 4 bytes + LockTime 4 bytes + TxContract 4 bytes + Serialized varint size for the
        # number of transaction inputs and outputs.
        n = 12 + VarIntSerializeSize(len(self.inputs)) + VarIntSerializeSize(len(self.outputs))
        for _input in self.inputs:
            n += _input.serialize_size()
        for _output in self.outputs:
            n += _output.serialize_size()
        return n


class AsimovScript:
    @classmethod
    def SignatureHash(cls, tx: Transaction, in_idx, sub_script, hash_type=script.SIGHASH_ALL) -> bytes:
        tx = deepcopy(tx)
        for txin in tx.inputs:
            txin.sig_script = b''
        tx.inputs[in_idx].sig_script = sub_script
        s = Web3.toBytes(hexstr=tx.to_hex())
        s += struct.pack(b"<i", hash_type)
        return bytes(Hash(s))

    @classmethod
    def _sign(cls, account: Account, hashbuf, hash_type) -> bytes:
        sec_key = CBitcoinSecret.from_bytes(bytes.fromhex(remove_0x_prefix(account.private_key)))
        sig_script_bytes: bytes = sec_key.sign(hashbuf)
        sig_script_bytes += bytes([hash_type & 0xff])
        public_key = Web3.toBytes(hexstr=account.public_key.decode())

        script_buffer = BytesIO()
        script_buffer.write(bytes([len(sig_script_bytes)]))
        script_buffer.write(sig_script_bytes)
        script_buffer.write(bytes([len(public_key)]))
        script_buffer.write(public_key)
        return script_buffer.getvalue()

    @classmethod
    def sign(cls, tx: Transaction, key_pair: Account, in_idx, sub_script, hash_type=script.SIGHASH_ALL) -> bytes:
        hashbuf = cls.SignatureHash(tx, in_idx, sub_script, hash_type)
        sign_hash = cls._sign(key_pair, hashbuf, hash_type)
        return sign_hash



import os
from solc import compile_files
from .data_type import SmartContract


class AsimovSolc:
    """
    The primary entry point for working with solidity compiler.
    """

    @classmethod
    def set_solidity_compiler(cls, compiler_path: str) -> None:
        """
        set solidity compiler path

        :param compiler_path: solidity compiler path

        :return: None

        .. code-block:: python

            >>> from asimov import AsimovSolc
            >>> AsimovSolc.set_solidity_compiler("/usr/local/bin/solc")
        """
        os.environ['SOLC_BINARY'] = compiler_path

    @classmethod
    def compile(cls, source_file: str, **kwargs) -> dict:
        """
        compile solidity source file

        :param source_file: source file path
        :param kwargs: reference to compile_files function in `py-solc <https://github.com/ethereum/py-solc>`_ library
        :type kwargs: dict

        :return: multiple compiled contract objects in dict type

        .. code-block:: python

            >>> from asimov import AsimovSolc
            >>> AsimovSolc.compile("/path/to/my/sources/example.sol")
            {'Example': {
                'abi': [{
                    'inputs': [],
                    'payable': False,
                    'stateMutability': 'nonpayable',
                    'type': 'constructor'
                }],
                'address': None,
                'bytecode': '6080604052348015600f57600080fd5b50603580601d6000396000f3006080604052600080fd00a165627a7a72305820bf199053a6eea79c7732c9211fc200781b170db435d118cbd86d2ac117e2fa360029',
                'source': 'pragma solidity ^0.4.25;\n'
                          '\n'
                          'contract Example {\n'
                          '    constructor() public {}\n'
                          '}\n'
                }
            }
        """
        with open(source_file) as f:
            source_code = f.read()
        compiled_objects = compile_files([source_file], output_values=('abi', 'bin', 'ast'), **kwargs)
        contracts = dict()
        for key in compiled_objects:
            contracts[key.split(':')[-1]] = SmartContract(
                source_code,
                compiled_objects[key]['abi'],
                compiled_objects[key]['bin']
            )
        return contracts





from .node import Node
from .solc import AsimovSolc
from .data_type import SmartContract, Tx
from .constant import ASCOIN, TxType


class Template:
    def __init__(self, node: Node):
        self.node = node

    def submit(self, source, template_name, chosen_contract, tx_fee_type=ASCOIN) -> Tx:
        """
        submit a new template to asimov blockchain and return the transaction object :class:`~asimov.data_type.Tx`

        :param source: smart contract source file path
        :param template_name: template name
        :param chosen_contract: contract name
        :param tx_fee_type: transaction fee type
        :return: the transaction object :class:`~asimov.data_type.Tx`

        .. code-block:: python

            >>> from asimov import Node, Template
            >>> node = Node("http://seed.asimov.tech", "0x98ca5264f6919fc12536a77c122dfaeb491ab01ed657c6db32e14a252a8125e3")
            >>> t = Template(node)
            >>> tx = t.submit("tutorial.sol", "template name1", "Tutorial")
            >>> tx
            [id: 8414ceb0d6db9c6418cd62c022168d961e69de80f662f0cdd99669f5954955ae]
            # template id
            >>> tx.id
            '8414ceb0d6db9c6418cd62c022168d961e69de80f662f0cdd99669f5954955ae'
        """
        compiled_contract: SmartContract = AsimovSolc.compile(source)[chosen_contract]
        tx_data = self.node.build_data_of_create_template(
            1, template_name, compiled_contract.bytecode, compiled_contract.abi, compiled_contract.source)
        return self.node.call_write_function(
            contract_tx_data=tx_data, call_type=TxType.TEMPLATE, tx_fee_type=tx_fee_type
        ).broadcast()

    def deploy_contract(self, template_id: str, constructor_arguments=None,
                        asset_value=0, asset_type=ASCOIN, tx_fee_type=ASCOIN) -> (Tx, str):
        """
        deploy a contract based on a given template id and
        return the address of the newly deployed contract on asimov blockchain and transaction object

        :param template_id: template id
        :param constructor_arguments: contract constructor arguments
        :param asset_value: asset value to send
        :param asset_type: asset type to send
        :param tx_fee_type: transaction fee type
        :return: the transaction object :class:`~asimov.data_type.Tx` and the address of new contract

        .. code-block:: python

            >>> from asimov import Node, Template, constant
            >>> node = Node("http://seed.asimov.tech", "0x98ca5264f6919fc12536a77c122dfaeb491ab01ed657c6db32e14a252a8125e3")
            >>> t = Template(node)
            >>> tx = t.submit("tutorial.sol", "template name1", "Tutorial")
            >>> assert tx.check() is constant.SUCCESS
            >>> t.deploy_contract(tx.id)
            [id: 2b7af42268f1ae098bc959d52c2c7d05af9cb44b2f2b6c8cfeb5c3afab941d5b]
        """
        if constructor_arguments is None:
            constructor_arguments = []
        contract_template = self.node.get_contract_template(key=template_id)
        tx_data = self.node.build_data_of_deploy_contract(contract_template, constructor_arguments)
        tx = self.node.call_write_function(
            contract_tx_data=tx_data, call_type=TxType.CREATE,
            asset_value=asset_value, asset_type=asset_type,
            tx_fee_type=tx_fee_type
        ).broadcast()
        address = self.node._calc_contract_address(tx.transaction.vin, tx.transaction.vout)
        return tx, address



import json
from typing import Union
from web3 import Web3
import eth_abi
from .data_type import EvmLog, EvmLogs


# https://codeburst.io/deep-dive-into-ethereum-logs-a8d2047c7371
class EvmLogParser:
    """
    The primary entry point for working with Asimov smart contract execution logs.
    """

    @classmethod
    def parse(cls, raw_log: Union[dict, list], abi: Union[dict, str]) -> EvmLogs:
        """
        Parse asimov vm execution log

        :param raw_log: contract execution log
        :param abi: contract abi object in json format
        :return: parsed log list
        """
        if isinstance(abi, str):
            abi = json.loads(abi)
        if type(raw_log) is dict:
            raw_logs = [raw_log]
        else:
            raw_logs = raw_log

        events = [e for e in abi if e['type'] == 'event']
        events_hash = {}
        for event in events:
            signature = Web3.sha3(text=f"{event['name']}({','.join([_input['type'] for _input in event['inputs']])})")
            events_hash[signature.hex()] = event
        logs = EvmLogs()
        for _raw_log in raw_logs:
            log = EvmLog(_raw_log)
            event_abi = events_hash[log.topics[0]]
            indexed_types = [e['type'] for e in event_abi['inputs'] if e['indexed']]
            indexed_names = [e['name'] for e in event_abi['inputs'] if e['indexed']]
            types = [e['type'] for e in event_abi['inputs'] if not e['indexed']]
            names = [e['name'] for e in event_abi['inputs'] if not e['indexed']]
            indexed_values = [eth_abi.decode_single(t, Web3.toBytes(hexstr=v)) for t, v in zip(indexed_types, log.topics[1:])]
            values = list(eth_abi.decode_abi(types, Web3.toBytes(hexstr=log.data)))
            logs.append({
                "name": event_abi['name'],
                "args": {n: v for n, v in zip(names + indexed_names, values + indexed_values)}
            })
        return logs



import json
import functools

from web3.utils.contracts import encode_transaction_data as __encode_transaction_data
from web3.utils.abi import filter_by_name, filter_by_type
from eth_abi import encode_abi

from asimov.data_type import Account


encode_transaction_data = functools.partial(__encode_transaction_data, None)


class AsimovJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Account):
            return {"private_key": obj.private_key, "public_key": obj.public_key, "address": obj.address}
        elif isinstance(obj, bytes):
            return obj.decode('utf-8')
        return super().default(obj)


def find_matching_func(contract_abi, fn_name, fn_type):
    """
    :param contract_abi:
    :param fn_name:
    :param fn_type:
    :return:
    {
        "inputs":[
            {
                "name":"_organizationName",
                "type":"string"
            },
            {
                "name":"_members",
                "type":"address[]"
            }
        ],
        "payable":true,
        "stateMutability":"payable",
        "type":"constructor",
        "name":"Association"
    }
    """
    if fn_name:
        abi = filter_by_name(fn_name, contract_abi)
    elif fn_type:
        abi = filter_by_type(fn_type, contract_abi)
    else:
        raise Exception("invalid parameters")

    if len(abi) == 0:
        raise Exception(f"no match function: {fn_name}")
    if len(abi) > 1:
        raise Exception(f"multiple match functions: {fn_name}")
    return abi[0]


def package_contract_func_args(contract_abi, fn_name, fn_type, args: list) -> list:
    """ setup parameter types according to abi"""
    abi = find_matching_func(contract_abi, fn_name, fn_type)

    for idx, input in enumerate(abi['inputs']):
        if input['type'].startswith('uint') or input['type'].startswith('int'):
            args[idx] = int(args[idx])
    return args


def encode_params(contract_abi, fn_name, fn_type, args: list) -> str:
    abi = find_matching_func(contract_abi, fn_name, fn_type)
    args = package_contract_func_args(contract_abi, fn_name, fn_type, args)
    types = [e['type'] for e in abi['inputs']]
    return encode_abi(types, args).hex()



import sys
import eth_utils


def uncache(exclude: [str]):
    pkgs = []
    to_uncache = []
    mod: str
    for mod in exclude:
        pkgs.append(mod.split('.', 1)[0])
    pkgs = list(set(pkgs))

    for mod in sys.modules:
        if mod in exclude:
            continue
        if mod in pkgs:
            to_uncache.append(mod)
            continue
        for pkg in pkgs:
            if mod.startswith(pkg + '.'):
                to_uncache.append(mod)
                break
    for mod in to_uncache:
        del sys.modules[mod]


def _is_hex_address(value) -> bool:
    """
    Checks if the given string of text type is an address in hexadecimal encoded form.
    """
    if not eth_utils.types.is_text(value):
        return False
    elif not eth_utils.hexadecimal.is_hex(value):
        return False
    else:
        unprefixed = eth_utils.hexadecimal.remove_0x_prefix(value)
        return len(unprefixed) == 42


def _is_checksum_address(value) -> bool:
    return True


from eth_utils import address
address.is_checksum_address = _is_checksum_address
address.is_hex_address = _is_hex_address
uncache(["eth_utils.address"])

from eth_abi.decoding import AddressDecoder
AddressDecoder.value_bit_size = 21 * 8




from setuptools import (
    setup,
    find_packages,
)

with open("README.md", "r") as f:
    long_description = f.read()


extras_require = {
    'test': [
        "pytest>=5.2.2,<6",
        "pytest-cov>=2.8.1,<3",
        "tox>=3.15.0,<4",
    ],
    'lint': [],
    'doc': [
        "Sphinx>=1.8.5,<2",
        "sphinx-rtd-theme>=0.4.3,<1",
    ],
    'dev': [
        "pytest-watch>=4.2.0,<5",
        "ipython",
        "wheel",
        "twine",
        "bumpversion>=0.5.3,<1",
    ]
}

extras_require['dev'] = (
    extras_require['dev'] +
    extras_require['test'] +
    extras_require['doc']
)

setup(
    name="py-asimov",
    version='0.3.0',
    keywords="asimov",
    description="""sdk for asimov chain""",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT License",
    author="The Asimov Foundation",
    author_email="ericsgy@163.com",
    include_package_data=True,
    url="https://gitlab.asimov.work/asimov/asimov-python-sdk",
    install_requires=[
        "fastecdsa>=1.7.4,<2",
        "web3>=4.9.0,<5",
        "python-bitcointx>=1.0.1,<2",
        "requests>=2.22.0,<3",
        "py-solc>=3.2.0,<4",
    ],
    python_requires=">=3.6, <4",
    extras_require=extras_require,
    zip_safe=False,
    py_modules=['asimov'],
    packages=find_packages(exclude=["test", "test.*"]),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.6',
    ],
)



__all__ = [
    "AccountFactory",
    "Address",
    "Transaction",
    "Contract",
    "Node",
    "AsimovSolc",
    "EvmLogParser",
    "Template",
    "constant",
    "Asset",
    "PrivateKeyFactory",
]

from ._monkey_patch import *

from asimov.account import (
    AccountFactory,
    Address,
    PrivateKeyFactory,
)
from asimov.transactions import Transaction
from asimov.contract import Contract
from asimov.node import Node
from asimov.solc import AsimovSolc
from asimov.evm_log import EvmLogParser
from asimov.template import Template
from asimov import constant
from asimov.data_type import Asset
